(* Title: BridgePIDE/template.sml
   Author: 

  1. The keyword \<open>Example\<close> shall insert the template underneath
  2. The template is filled from hidden data:
    2.1. \<open>Model\<close>:
        The elements come with hints for input the format,
        see test/../i-model.sml --- build I_Model.init_TEST ---
    2.2 \<open>References\<close>:
        * The elements come with hints for input the format, see test/../Test_VSCode_Example.thy
        * toggle \<Odot>\<Otimes> determines whether the \<open>Model\<close> relates to \<open>Problem_Ref\<close> or \<open>Method_Ref\<close>
  3. Input to 2.1 is 
    * checked and marked with feedback
    * datatype feedback = Correct | Incomplete | SyntaxError | Superfluous | False
  4. Input to 2.2 eventually by selection from a tree (a DAG for theories)
*)

signature TEMPLATE =
sig
  datatype feedback = Correct | Incomplete | SyntaxError | Superfluous | False
  val show: Proof.context -> (bool * term) list -> I_Model.T_TEST -> References.input -> unit

(*from isac_test for Minisubpbl*)
  type model_out
  val model_to_string: model_out -> string
  val refs_to_string: References.input -> string
  val model_to_out: Proof.context -> (bool * term) list -> I_Model.T_TEST -> model_out
  val items_to_out: Proof.context -> Model_Pattern.m_field -> I_Model.T_TEST ->
    Model_Pattern.m_field * (UnparseC.term_as_string * feedback) list
  val item_to_out: Proof.context -> I_Model.single_TEST -> UnparseC.term_as_string * feedback

\<^isac_test>\<open>
(**)
\<close>
end

(**)
structure Template(**) : TEMPLATE(**) =
struct
(**)

datatype feedback = Correct | Incomplete | SyntaxError | Superfluous | False
type item_out = UnparseC.term_as_string * feedback
type field_out = Model_Pattern.m_field * item_out list
type model_out =
    field_out (* "#Given", feedback: Correct | Incomplete | SyntaxError | Superfluous *)
  * field_out (* "#Where", feedback: Correct | False                                  *)
  * field_out (* "#Given", feedback: Correct | Incomplete | SyntaxError | Superfluous *)
  * field_out (* "#Given", feedback: Correct | Incomplete | SyntaxError | Superfluous *)


(** create strings **)

fun feedback_to_string Correct = " (Correct)"
  | feedback_to_string Incomplete = " (Incomplete)"
  | feedback_to_string SyntaxError = " (SyntaxError)"
  | feedback_to_string Superfluous = " (Superfluous)"
  | feedback_to_string False = " (False)"
fun elem_to_string false (term_as_string, fb) =
    quote term_as_string ^ feedback_to_string fb ^ "\n"
  | elem_to_string true (term_as_string, fb) =
    "        " ^ quote term_as_string ^ feedback_to_string fb ^ "\n"
fun field_to_string (m_field, []) = m_field ^ ": \n"
  | field_to_string (m_field, elem :: elems) =
    m_field ^ ": " ^ elem_to_string false elem ^ 
    fold (curry op ^) (map (elem_to_string true) (elems)) ""
fun model_to_string (given, where_, find, relate) =
  "  Specification:\n" ^
  "    Model:\n" ^
  "      " ^ field_to_string given ^
  "      " ^ field_to_string where_ ^
  "      " ^ field_to_string find ^
  "      " ^ field_to_string relate (*^
  "    References:\n"*)


(** create output format **)

fun item_to_out ctxt (_, _, _, _, (I_Model.Cor_TEST dsc_ts, _)) =
    (dsc_ts |> Term.list_comb |> UnparseC.term ctxt, Correct)
  | item_to_out _ (_, _, _, _, (I_Model.Syn_TEST (term_as_string), _)) =
    (term_as_string, Incomplete)

  | item_to_out ctxt (_, _, _, _, (I_Model.Inc_TEST (dsc, []), _)) =
    (UnparseC.term ctxt dsc ^ " " ^ Model_Pattern.empty_for dsc, Incomplete)
  | item_to_out ctxt (_, _, _, _, (I_Model.Inc_TEST (dsc_ts), _)) =
    (dsc_ts |> Term.list_comb |> UnparseC.term ctxt, Incomplete)

  | item_to_out ctxt (_, _, _, _, (I_Model.Sup_TEST dsc_ts, _)) =
    (dsc_ts |> Term.list_comb |> UnparseC.term ctxt, Superfluous)

fun items_to_out ctxt m_field items =
  (m_field,
    filter (fn (_, _, _, f, _) => f = m_field) items |> map (item_to_out ctxt))
fun where_to_out ctxt (bool, t) =
  (UnparseC.term ctxt t, if bool then Correct else False)
fun wheres_to_out ctxt wheres = ("#Where", map (where_to_out ctxt) wheres)
fun model_to_out ctxt where_ model =
  let
    val given = items_to_out ctxt "#Given" model
    val where_ = wheres_to_out ctxt where_
    val find = items_to_out ctxt "#Find" model
    val relate = items_to_out ctxt "#Relate" model
  in (given, where_, find, relate) end
fun refs_to_string (thy, pbl, met) =
  "    References:\n" ^
  "      Theory_Ref: " ^ quote thy ^ "\n" ^
  "      Problem_Ref: " ^ quote pbl ^ "\n" ^
  "      Method_Ref: " ^ quote met ^ "\n" ^
  "  Solution:"


(** a fake version for Template.show **)

fun show ctxt precond i_model in_refs =
  let
    val model_out = model_to_out ctxt precond i_model
    val fake_string = model_to_string model_out ^ refs_to_string in_refs
    val _ = writeln fake_string
  in () end

(**)end(*struct*);
