(* Title:  MathEngBasic/calc-tree-elem.sml
   Author: Walther Neuper 191026
   (c) due to copyright terms

Definitions for Tactic.T and Ctree.
*)
signature MODEL_DEFINITION =
sig

  type form_T
  type form_model
  type form_refs
  val form_empty : form_T

  type variant
  type variants

  type m_field = Model_Pattern.m_field
  type descriptor
  val is_list_descr: descriptor -> bool
  type values
  val make_values: descriptor * term -> term list
  val values_to_present: values -> term

  type o_model
  type o_model_single
  val o_model_to_string : Proof.context -> o_model -> string
  val o_model_single_to_string : Proof.context -> o_model_single -> string
  val o_model_empty : o_model_single

  type i_model
  type i_model_TEST
  type i_model_single
  type i_model_single_TEST
  datatype i_model_feedback = 
      Cor of (term * (term list)) * (term * (term list))
    | Syn of TermC.as_string 
    | Typ of TermC.as_string 
    | Inc of (term * (term list))	* (term * (term list))
    | Sup of (term * (term list)) 
    | Mis of (term * term) 
    | Par of TermC.as_string
  datatype i_model_feedback_TEST = 
      Cor_TEST of (descriptor * values)
    | Syn_TEST of TermC.as_string 
    | Inc_TEST of (descriptor * values)
    | Sup_TEST of (descriptor * values) 
  val feedback_empty_TEST: i_model_feedback_TEST

  val i_model_empty : i_model_single
  val i_model_empty_TEST: i_model_single_TEST

  val get_descr_opt: i_model_single_TEST -> descriptor option;
  val get_dscr_opt: i_model_feedback_TEST -> descriptor option
  val get_descr: i_model_feedback_TEST -> descriptor
  val descriptor_exists: descriptor -> i_model_feedback_TEST -> bool

  val member_vnt: variants -> variant -> bool
  val max_variants: o_model -> i_model_TEST -> variants

(*from isac_test for Minisubpbl*)
  val all_variants: ('a * variants * 'c * 'd * 'e) list -> variants
  val filter_variants': i_model_TEST -> variant -> i_model_TEST
  val cnt_corrects: i_model_TEST -> int
  val arrange_args: int list -> int * variant list -> (int * variant) list

\<^isac_test>\<open>
(**)
\<close>

end

(**)                   
structure Model_Def(**): MODEL_DEFINITION(**) =        
struct
(**)

type m_field = Model_Pattern.m_field;
type descriptor = term;
type values = term list;
(*
Isac implements list with relaxed representation of types to the user
and holds list elements in a specific format in order to simplify element-wise input.

Legend: lists in ML                   [[______________________]]   [[______________________]]
        lists in HOL                      [________________]           [________________]
                                                                  
    descriptor      Formalise.model       O_Model                      I_Model
                    _ $ term              values                       values
(1) equalities      eqs               [[  eqs                 ]]   [[  eqs                 ]]
(2) equalities      [c_2 = 0, L *..]  [[  [c_2 = 0], [L *..]  ]]   [[  [c_2 = 0], [L *..]  ]]
(3) solveForVars    [c, c_2, c_3]     [[  [c], [c_2], [c_3]   ]]   [[  [c], [c_2], [c_3]   ]]
(4) Constants       [r = 7]           [[  [[r = 7]]           ]]   [[  [[r = 7]]           ]]
(5) Maximum         A                 [[  A                   ]]   [[  A                   ]]
                              ---> fun make_values (descr, ts) -->
                              <--- fun make_present (descr, ts) <--

Observe the elements of HOL-lists enclosed in ML-lists in lines (2..4).
Without those (strange!) HOL-lists in values one could not distinguish (1) and (3), 
  in case ony one element is input ("Inc (solveForVars c)")
  and the representation to the user could not distinguish between "c" and "[c]";
  then stepwise input would present suprprises with "c_2 = 0" .. "[c_2 = 0, L *..]" .
The choice for this design of handling "[]" allows to distinguis (1) from (2,3) only from
  values without referring to descriptor.

For descriptors (1..4) Model_Def.is_list_descr;
for (5) "Maximum" we have is_list_descr = false, for non-lists the above distinction is irrelevant.
*)

fun is_list_descr descr =
  case Model_Pattern.descr_type (type_of descr) of Model_Pattern.List _ => true | _ => false
fun make_values (descr, ts) = 
  if is_list_descr descr
  then if TermC.is_list ts
    then ts |> TermC.isalist2list |> map TermC.wrap_HOL_list
    else [ts]
  else [ts]
fun values_to_present values =
  if TermC.is_list (hd values)
    then values |> map TermC.isalist2list |> flat |> TermC.list2isalist'
    else hd values


(** definitions for Formalise.T **)

type form_model = TermC.as_string list;
type form_refs = References_Def.T;
type form_T = form_model * form_refs;
val form_empty = ([], References_Def.empty);

type variant = int;
type variants = variant list; (* variants given by Formalise.T *)

(** definitions for I_Model.T **)

datatype i_model_feedback = 
  Cor of (descriptor * (* correct input: descriptor, see Input_Descript.thy *)
     (values)) *              (* set of values for list-types: for decdiing
                                 Inc the values packed out from the list;
                                 They are packed back to an isalist in env  *)
   (term * (term list))       (* will be eliminated in i_model_feedback_TEST*)
| Syn of TermC.as_string      (* PIDE: contains "(__=__)" etc               *)
| Typ of TermC.as_string      (* type error: drop with PIDE ?               *)
| Inc of (descriptor * values) * (term * (term list))  (* see Cor           *)
| Sup of (descriptor * values)(* superfluous; input not found in O_Model.T  *)
| Mis of (descriptor * term)  (* missing; for feedback                      *)
| Par of TermC.as_string;     (* internal state from fun parsitm            *)

(*
  this datatype assumes that all items of the Model_Pattern are output
  by the initial template, where the items are displayed with hints for the input format.
*)
datatype i_model_feedback_TEST = 
  Cor_TEST of (descriptor * (* correct input: descriptor, see Input_Descript.thy *)
     (values))                (* set of values for list-types: for decding
                                 Inc the values packed out from the list;
                                 They are packed back to an isalist for envs     *)
| Syn_TEST of TermC.as_string (* syntax error kept for P_Model.switch_pbl_met    *)
| Inc_TEST of (descriptor * values) (* see Cor                                   *)
| Sup_TEST of (descriptor * values) (* superfluous; input not found in O_Model.T *)
val feedback_empty_TEST = Syn_TEST "i_model_empty"


type i_model_single = 
  int *             (* sequence of input, 0 = untouched, drop with PIDE ?  *)
  variants *        (* pointers to variants given in Formalise.model       *)
  bool *            (* input on this item is not/complete                  *)
  m_field *         (* #Given | #Find | #Relate                            *)
  i_model_feedback; (* variables and values, see above                     *)
type i_model_single_TEST = 
  int *                     (* sequence of input, 0 = untouched, drop with PIDE ? *)
  variants *                (* pointers to variants given in Formalise.model      *)
  bool *                    (* this item is not/complete, completed by O_Model.T  *)
  m_field *                 (* #Given | #Find | #Relate                           *)
  (i_model_feedback_TEST *  (* variables and values, see above                    *)
    Position.T);            (* for pushing feedback back to PIDE                  *)

type i_model = i_model_single list;
type i_model_TEST = i_model_single_TEST list;

(*shift to i-model.sml ?*)
val i_model_empty = (0, [], false, "i_model_empty", Syn "i_model_empty");
val i_model_empty_TEST = (0, [], false, "i_model_empty", (feedback_empty_TEST, Position.none));

(*get descriptor of those items which can contribute to Subst.T and Env.T*)
fun get_dscr_opt (Cor_TEST (descr, _)) = SOME descr
  | get_dscr_opt (Inc_TEST (descr, _)) = SOME descr
  | get_dscr_opt (Sup_TEST (descr, _)) = SOME descr
  | get_dscr_opt _ = NONE (*other feedback cannot contribute to make_environments *)
fun get_descr_opt (_, _, _, _, (feedb, _)) =
  case get_dscr_opt feedb of NONE => NONE | some_descr => some_descr
fun descriptor_exists descr feedb =
  case get_dscr_opt feedb of
    SOME descr' => descr' = descr
  | NONE => raise ERROR "existing_description NONE"
fun get_descr feedb = case get_dscr_opt feedb of
    SOME descr => descr
  | NONE => raise ERROR "get_descr_opt from item without this description"  


(** max_variants **)

fun member_vnt [] _ = true
  | member_vnt vnts vnt = member op= vnts vnt

fun all_variants model =
  map (fn (_, variants, _, _, _) => variants) model
  |> flat
  |> distinct op =
fun filter_variants' i_singles n = 
  filter (fn (_, variants, _, _, _) => member op= variants n) i_singles

fun some_input (Cor_TEST _) = true
  | some_input (Inc_TEST (_, _::_)) = true
  | some_input (Syn_TEST _) = true
  | some_input _ = false

fun cnt_corrects i_model = 
  fold (curry op +) (map (fn (_, _, _, _, (feedb, _)) =>
    if some_input feedb then 1 else 0) i_model) 0;
fun arrange_args [] _ = []
  | arrange_args (sum :: ss) (cnt, all) = (sum, nth cnt all) :: (arrange_args ss (cnt + 1, all)) 
fun max_variants o_model i_model =
  let
    val all_variants =
        map (fn (_, variants, _, _, _) => variants) i_model
        |> flat
        |> distinct op =
    val variants_separated = map (filter_variants' i_model) all_variants
    val sums_corr = map (cnt_corrects) variants_separated
    val sum_variant_s = arrange_args sums_corr (1, all_variants)

    val max_first = rev (sort (fn ((i, _), (j, _)) => int_ord (i, j)) sum_variant_s)
    val maxes = filter (fn (cnt, _) => curry op = (fst (hd max_first)) cnt) max_first
      |> map snd
  in
    if maxes = []
    then map (fn (_, variants, _, _, _) => variants) o_model
        |> flat
        |> distinct op =
    else maxes
  end


(** definitions for O_Model.T **)

type o_model_single =
  (int *       (* sequence of input, 0 = untouched, drop with PIDE *)
	variants *   (* pointers to variants given in Formalise.T        *)
	m_field *    (* #Given | #Find | #Relate + TODO: #Where          *)
	descriptor * (* see Input_Descript.thy                           *)
	values       (* isalist2list t | [t]: the values of in item      *)
	);
type o_model = o_model_single list;  
val o_model_empty = (0, [], "", TermC.empty, [TermC.empty]) : o_model_single;

fun o_model_single_to_string ctxt (i, vs, fi, t, ts) = 
  "(" ^ string_of_int i ^ ", " ^ (strs2str o (map string_of_int)) vs ^ ", " ^ fi ^ ", " ^
  UnparseC.term ctxt t ^ ", " ^ (strs2str o (map (UnparseC.term ctxt))) ts ^ ")";
fun o_model_to_string ctxt o_model =
  (strs2str' o (map (linefeed o (o_model_single_to_string ctxt)))) o_model;


(**)end(**)
