(* Title:  Specify/specification.sml
   Author: Walther Neuper, Mathias Lehnfeld
   (c) due to copyright terms
*)

(* Survey on type fmz_ .. type ori .. type itm (in probl, meth) .. formal args of root-/SubProblem
   and relations between respective list elements:                                       #1#
                  fmz      =             [ dsc $ v......................................(dsc $ v)..]
root problem on pos = ([], _)            
                  fmz_     =             [(dsc, v).......................................(dsc, v)..]
                  \<down>                                                                       
                  oris     =             [(dsc, v)..(dsc, v),(dsc, v),(dsc, v)...........(dsc, v)..]
                  \<down>                       #Given,..,#Relate  #Find     #undef............#undef   \<down>
                  \<down>                                                     \<down>                 \<down>       \<down>
  Specify_Problem + pbl.model=             [(dsc,id)..(dsc,id),(dsc,id)]  \<down>                 \<down>       \<down>
                  \<down>                                                     \<down>                 \<down>       \<down>
                  itms     =             [(dsc, v)..(dsc, v),(dsc, v)]  \<down>                 \<down>       \<down>
  int.modelling                           +Cor ++++++++++Cor +Cor       \<down>                 \<down>       \<down>
                  \<down>                                                     \<down>                 \<down>       \<down>
  Specify_Method  + met.model=             [(dsc,id)..(dsc,id),(dsc,id)]  \<down>                 \<down>       \<down>
                  \<down>                                                     \<down>                 \<down>       \<down>
                  itms     =             [(dsc, v)..(dsc, v),(dsc, v),(dsc, v)..(dsc, v)] \<down>       \<down>
  int.modelling                                                       +Cor ++++++Cor      \<down>       \<down>
                  form.args=             [ id ..... id      , ///////  id ....... id    ] \<down>       \<down>
                  env =                  [(id, v)..(id, v)  , /////// (id, v)....(id, v)] \<down>       \<down>
  interpret code env = [(id, v)..(id, v), (id, v)..(id, v)  , /////// (id, v)....(id, v)] \<down>       \<down>
    ..extends env       ^^^^^^^^^^^^^^^^   \<down>                           \<down>   \<down>              \<down>       \<down>
                                   \<down>       \<down>                           \<down>   \<down>              \<down>       \<down>
SubProblem on pos = ([2], _)       \<down>       \<down>                           \<down>   \<down>              \<down>       \<down>
                  form.args=      [id ................................ id ]\<down>              \<down>       \<down>
                  \<down>                REAL..BOOL..                            \<down>              \<down>       \<down>
                  \<down>                                                        \<down>              \<down>       \<down>
                  + met.model=      [(dsc,id).......................(dsc,id)]\<down>              \<down>       \<down>
                                    \<down>                               \<down>      \<down>              \<down>       \<down>
                  oris     =      [(dsc, v)...................... (dsc,   v) ........... (dsc, v)]\<down>
  Specify_Problem, int.modelling, Specify_Method, interpret code as above                 #1#    \<down>
                                                                                                  \<down>
SubProblem on pos = ([3, 4], _)                                                                   \<down>
                  form.args=              [id ...................... id ]                         \<down>
                  \<down>                        REAL..BOOL..                                           \<down>
                  + met.model=              [(dsc,id).............(dsc,id)]                         \<down>
                  oris     =              [(dsc, v).............(dsc, v)...................(dsc, v)] 
  Specify_Problem, int.modelling, Specify_Method, interpret code as above                    #1#

Notes:
(1) SubProblem implements sub-function calls such, that the arguments(in/output + where_- + post-cond.)
    of the functions become concern of interactive modelling.
(2) In Isac-terms, the above concerns the phases of modelling and 
    and of specifying (Specify_Problem, Specify_Method),
    while stepwise construction of Solution is called solving.
    The latter is supported by Lucas-Interpretation of the functions' body.
(3) ori list is determined by fmz_ (in root-pbl) or by args of SubProblem;
    it is as complete as possible (this has been different up to now).
(4) itm list is initialised by pbl-model | met-model and completed (Cor) by stepwise user input.
(5) model-pattern of pbl can omit technical items, e.g. boundVariable or functionName,
    add them to the model-pattern of met and let this input be done automatically;
    respective items must be in fmz.
*)
signature SPECIFICATION =
sig

  type T = Specification_Def.T (*see NOTES AT START PIDE-ADAPTION in Specification_Def*)
  val get: Calc.T -> T
  val get_data: Calc.T ->
    I_Model.T * O_Model.T * References.T * I_Model.T * References.T * Proof.context
  val reset: Calc.T -> Calc.T

  val is_complete': I_Model.T -> Pre_Conds.T -> References.T -> bool(*TODO:see I_Model.is_complete*)
  val is_complete: Calc.T -> bool                                   (*TODO:see I_Model.is_complete*)
end

(**)
structure SpecificationC(**): SPECIFICATION(**) =
struct
(**)

type T = Specification_Def.T;

(** is SpecificationC complete? **)

(* check i_model either for Problem or MethodC *)
(*whether 
(1) singles are all I_Model.Cor 
(2) Pre_Conds are true
 * TODO-PIDE:
   * replace (1..2) by I_Model.is_complete
   * sort out References
 *)
fun is_complete' its where_ (dI, pI, mI) = 
  foldl and_ (true, map #3 (its: I_Model.T)) andalso 
  foldl and_ (true, map #1 (where_: Pre_Conds.T)) andalso 
  dI <> ThyC.id_empty andalso pI <> Problem.id_empty andalso mI <> MethodC.id_empty

fun is_complete (pt, (p, _)) =
  let
    val (itms, oris, probl, o_spec, spec, ctxt) = case Ctree.get_obj I pt p of
       Ctree.PblObj {meth = itms, origin = (oris, o_spec, _), probl, spec, ctxt, ...}
       => (itms, oris, probl, o_spec, spec, ctxt)
    | _ => raise ERROR "SpecificationC.is_complete only with PblObj"
    val (_, pbl_id, met_id) = References.select_input o_spec spec
  in
    I_Model.s_are_complete ctxt oris
        (I_Model.OLD_to_TEST probl, I_Model.OLD_to_TEST itms) (pbl_id, met_id)
  end

(** handle acces to Ctree **)

fun get_data (pt, (p, _)) =
  case Ctree.get_obj I pt p of
    (Ctree.PblObj {meth, origin = (oris, (dI', pI', mI'), _), probl, spec = (dI, pI, mI), ctxt, ...})
      => (meth, oris, (dI', pI', mI'), probl, (dI, pI, mI), ctxt)
  | _ => raise ERROR "specify_additem: uncovered case of get_obj I pt p";

(* get a calchead from a PblObj-node in the ctree; preconditions must be calculated *)
fun get (pt, (p, Pos.Pbl)) = 
    let
	    val (ospec, hdf', spec, probl, ctxt) = case Ctree.get_obj I pt p of
	      Ctree.PblObj {origin = (_, ospec, hdf'), spec, probl, ctxt, ...} => (ospec, hdf', spec, probl, ctxt)
	    | _ => raise ERROR "get Pbl: uncovered case get_obj"
      val {where_rls, where_, model, ...} =
        Problem.from_store ctxt (#2 (References.select_input ospec spec))
      val (_, where_) = Pre_Conds.check ctxt where_rls where_ 
        (model, I_Model.OLD_to_TEST probl)
    in
      (is_complete' probl where_ spec, Pos.Pbl, hdf', probl, where_, spec)
    end
  | get (pt, (p, Pos.Met)) = 
    let
		  val (ospec, hdf', spec, meth, ctxt) = case Ctree.get_obj I pt p of
		    Ctree.PblObj {origin = (_, ospec, hdf'), spec, meth, ctxt, ...} => (ospec, hdf', spec, meth, ctxt)
		  | _ => raise ERROR "get Met: uncovered case get_obj"
      val {where_rls, where_, model, ...} =
        MethodC.from_store ctxt (#3 (References.select_input ospec spec))
      val (_, where_)  = Pre_Conds.check ctxt where_rls where_ 
        (model, I_Model.OLD_to_TEST meth)
    in
      (is_complete' meth where_ spec, Pos.Met, hdf', meth, where_, spec)
    end
  | get _ = raise ERROR "get: uncovered definition"

(* at the activeFormula set the Specification 
   to empty and return a CalcHead;
   the 'origin' remains (for reconstructing all that) *)
fun reset (pt, (p,_)) = 
  let
    val () = case Ctree.get_obj I pt p of
      Ctree.PblObj _ => ()
    | _ => raise ERROR "reset: uncovered case get_obj"
    val pt = Ctree.update_pbl pt p []
    val pt = Ctree.update_met pt p []
    val pt = Ctree.update_spec pt p References.empty
  in (pt, (p, Pos.Pbl)) end

(**)end(**)
